UNOManager = UNOManager or class()

function UNOManager:init()
	if not PackageManager:loaded("packages/start_menu") then
		PackageManager:load("packages/start_menu")
	end

	self._arena_info = tweak_data.uno:get_arena(4)

	self._turn = 0
	self._turn_order = 1

	self._skip_count = 0
	self._draw_count = 0

	self._player_to_seat = {}
	self._characters = {}

	self:_setup_piles()
	self:_setup_hands()
	self:_setup_selectors()

	self:_register_network_messages()
end

function UNOManager:_setup_piles()
	if self._draw_pile then
		self._draw_pile:destroy()
	end

	self._draw_pile = UNOCardPile:new(
		self._arena_info.draw_pile[1],
		self._arena_info.draw_pile[2],
		true,
		true
	)

	self:spawn_card():set_owner(self._draw_pile)

	if self._play_pile then
		self._play_pile:destroy()
	end

	self._play_pile = self._play_pile or UNOCardPile:new(
		self._arena_info.play_pile[1],
		self._arena_info.play_pile[2],
		false,
		false
	)
end

function UNOManager:_setup_hands()
	if self._hands then
		for _, hand in pairs(self._hands) do
			self._hands:destroy()
		end
	end

	self._hands = {}

	for index = 1, #self._arena_info.cameras do
		local location = self._arena_info.cameras[index]
		self._hands[index] = UNOHand:new(location[1], location[2])
	end
end

function UNOManager:local_hand()
	return self._hands[managers.network:session():local_peer():id()]
end

function UNOManager:_setup_selectors()
	if self._selectors then
		for _, selector in pairs(self._selectors) do
			self._selectors:destroy()
		end
	end

	self._selectors = {}

	for index = 1, #self._arena_info.cameras do
		local location = self._arena_info.cameras[index]
		self._selectors[index] = UNOCardSelector:new(location[1], location[2])
	end
end

function UNOManager:local_selector()
	return self._selectors[managers.network:session():local_peer():id()]
end

function UNOManager:set_turn(turn)
	self._turn = turn
end

function UNOManager:is_turn(turn)
	return self._turn == turn
end

function UNOManager:is_hand_turn(hand)
	for index, potential_hand in pairs(self._hands) do
		if potential_hand == hand then
			return self._turn == index
		end
	end

	return false
end

function UNOManager:flip_turn_direction()
	self._turn_order = self._turn_order * -1
end

function UNOManager:set_skip_count(count)
	self._skip_count = count
end

function UNOManager:set_draw_count(count)
	self._draw_count = count
end

function UNOManager:player_count()
	return #self._player_to_seat
end

function UNOManager:_validate_turn()
	self._turn = ((self._turn - 1) % self:player_count() + 1)
end

function UNOManager:_ai_play_card(player_id, card, card_index)
	card:set_owner(self._play_pile)

	local default_behaviour_overriden = false
	local ai_placement_clbk = card:ai_placement_clbk()
	if ai_placement_clbk then
		default_behaviour_overriden = ai_placement_clbk(player_id, card, card_index)
	end

	if not default_behaviour_overriden then
		managers.uno_network:send_peers_except_request(1, "SYNC_PLAY_CARD", {player_id, card:card_id(), card_index})
		managers.menu_component:post_event("loot_flip_card")
	end

	local server_placement_clbk = card:server_placement_clbk()
	if server_placement_clbk then
		server_placement_clbk()
	end
end

function UNOManager:_server_draw_card(player_id, force_play)
	local peers_table = managers.network and managers.network:session() and managers.network:session():all_peers()
	local is_player = peers_table and peers_table[player_id]
	local card_id = tweak_data.uno:get_random_card_id()
	local card_tweak_data = tweak_data.uno.cards[card_id]
	local turn_finished = true

	if is_player then
		if force_play and self:can_play_combination(card_tweak_data.type, card_tweak_data.value) then
			managers.uno_network:send_peer_request(player_id, "CLIENT_FORCE_PLAY", {card_id})
			turn_finished = false
		else
			managers.uno_network:send_peer_request(player_id, "CLIENT_GIVE_CARD", {tweak_data.uno.to_obfuscated[card_id]})
			managers.uno_network:send_peers_except_request(player_id, "SYNC_DRAW_CARD", {player_id})
		end
	else
		local hand = self._hands[player_id]
		local card = self:spawn_card()

		card:set_card_id(card_id)
		card:set_owner(hand)

		managers.menu_component:post_event("loot_flip_card")

		if force_play and self:can_play_combination(card_tweak_data.type, card_tweak_data.value) then
			self:_ai_play_card(player_id, card, hand:card_amount())
		else
			managers.uno_network:send_peers_except_request(1, "SYNC_DRAW_CARD", {player_id})
		end
	end

	return turn_finished
end

function UNOManager:increment_turn()
	self._turn = self._turn + self._turn_order
	self:_validate_turn()

	if self._draw_count > 0 then
		for i = 1, self._draw_count do
			self:_server_draw_card(self._turn, false)
		end

		self._draw_count = 0
	end

	if self._skip_count then
		self._turn = self._turn + (self._skip_count * self._turn_order)
		self:_validate_turn()

		self._skip_count = 0
	end
end

function UNOManager:can_play_card(card)
	return self:can_play_combination(card:type(), card:value())
end

function UNOManager:can_play_combination(type, value)
	if not self._play_pile then return end

	local top_card = self._play_pile:top_card()
	if not top_card then return true end

	local placement_rule = top_card:placement_rule()
	if not placement_rule then return true end

	local placement_function = tweak_data.uno.placement_rules[placement_rule]
	if not placement_function then return true end

	return placement_function(top_card:type(), top_card:value(), type, value)
end

function UNOManager:setup_controllers()
	local peers_table = managers.network and managers.network:session() and managers.network:session():all_peers()
	local peers_count = table.size(peers_table or {})
	if peers_count == 0 then return end

	for peer_id, _ in pairs(peers_table) do
		self._player_to_seat[peer_id] = peer_id
	end

	local local_peer_id = managers.network:session():local_peer():id()
	local camera_location = self._arena_info.cameras[local_peer_id]
	self._controller = UNOController:new(local_peer_id, camera_location[1], camera_location[2])

	self._controller:add_controllable(self._draw_pile)
	self._controller:add_controllable(self._hands[local_peer_id])
	self._controller:add_controllable(self._selectors[local_peer_id])

	local ai_count = #self._arena_info.cameras - peers_count
	if not managers.groupai:state():team_ai_enabled() then
		ai_count = 0

		if peers_count < 2 then
			ai_count = 1
		end
	end

	for ai_index = 1, ai_count do
		local player_index = peers_count + ai_index
		self._player_to_seat[player_index] = player_index
	end
end

function UNOManager:controller()
	return self._controller
end

function UNOManager:remove_controllers()
	if self._controller then
		self._controller:destroy()
		self._controller = nil
	end
end

function UNOManager:spawn_card(position, rotation)
	position = position or self._arena_info.draw_pile[1]
	rotation = (rotation or self._arena_info.draw_pile[2]) * Rotation(0, -90, 0)

	local card = UNOCard:new(position, rotation)

	return card
end

function UNOManager:_handle_ai_control(t, dt, player_id)
	if not self:is_turn(player_id) then return end

	local hand = self._hands[player_id]

	local decision_time = tweak_data.uno.ai_parameters.decision_time
	local decision_time_random = tweak_data.uno.ai_parameters.decision_time_random

	if not self._turn_play_t then
		self._turn_play_t = t + decision_time + (math.random() * decision_time_random)
		hand:update_cards()
	end

	if self._turn_play_t and self._turn_play_t > t then return end

	for index, card in pairs(hand:cards()) do
		if not card:card_id() then
			local card_id = tweak_data.uno:get_random_card_id()
			card:set_card_id(card_id) 
		end
	end

	local played_card = false
	for index, card in pairs(hand:cards()) do
		if managers.uno:can_play_card(card) then
			self:_ai_play_card(player_id, card, index)

			played_card = true

			break
		end
	end

	if not played_card then
		self:_server_draw_card(player_id, true)
	end

	self:increment_turn()
	managers.uno_network:send_peers_request("SYNC_TURN", {self._turn})

	hand:update_cards()
	self._turn_play_t = nil
end

function UNOManager:_spawn_characters()
	local peers_table = managers.network and managers.network:session() and managers.network:session():all_peers()
	local peers_count = table.size(peers_table or {})
	if peers_count == 0 then return end

	for peer_id, peer in pairs(peers_table) do
		self._player_to_seat[peer_id] = peer_id

		local character_location = self._arena_info.characters[peer_id]

		local character_name = peer:character()
		managers.criminals:character_by_name(character_name).taken = true

		local complete_outfit = peer:blackmarket_outfit()
		complete_outfit.character = character_name
		complete_outfit.visual_seed = peer._visual_seed
		complete_outfit.armor_id = peer._equipped_armor_id
		local character_id = managers.blackmarket:get_character_id_by_character_name(character_name)

		self._characters[peer_id] = UNOCharacter:new(character_location[1], character_location[2], character_id, complete_outfit)
	end

	local ai_count = #self._arena_info.cameras - peers_count
	if not managers.groupai:state():team_ai_enabled() then
		ai_count = 0

		if peers_count < 2 then
			ai_count = 1
		end
	end

	for ai_index = 1, ai_count do
		local player_index = peers_count + ai_index
		self._player_to_seat[player_index] = player_index

		if Network:is_server() then
			local character_name = managers.criminals:get_free_character_name()
			managers.criminals:character_by_name(character_name).taken = true
			local henchman_string = managers.blackmarket:henchman_loadout_string(ai_index)

			character_name = CriminalsManager.convert_old_to_new_character_workname(character_name)
			managers.uno_network:send_peers_request("SYNC_AI_CHARACTER", {player_index, character_name, henchman_string})
		end
	end
end

function UNOManager:update(t, dt)
	if self._controller then
		self._controller:update(t, dt)
	end

	if self._intro_started then
		self._intro_start_t = t
		self._current_intro_event = 0
		self._intro_started = false

		self:_spawn_characters()
	end

	for player_id, character in pairs(self._characters) do
		local character_location = self._arena_info.characters[player_id]
		character:set_position(character_location[1])
		character:set_rotation(character_location[2])
	end

	if self._intro_start_t then
		local intro_t = t - self._intro_start_t
		local intro_event = nil
		local intro_event_index

		for index, event_data in pairs(tweak_data.uno.intro_events) do
			local time_start = event_data[1]
			local time_end = event_data[2]
			local event = event_data[3]

			if self._current_intro_event ~= index and intro_t > time_start and intro_t < time_end then
				self._current_intro_event = index

				managers.dialog:queue_dialog(event, {
					position = self:local_hand()._position,
					skip_idle_check = Application:editor()
				})

				if index == #tweak_data.uno.intro_events then
					self._intro_end_t = t + (time_end - time_start)
				end

				break
			end

			if self._intro_end_t and self._intro_end_t <= t then
				self._intro_start_t = nil
				self._intro_end_t = nil

				if Network:is_server() then
					self._next_card_draw_t = t
					self._cards_to_distribute = tweak_data.uno.rules.starting_hand_size * self:player_count()
				end
			end
		end
	end

	if self._game_over then
		self._game_over = false

		self._death_t = t + tweak_data.uno.death_timer
	end

	if self._death_t and t >= self._death_t then
		managers.menu_component:post_event(tweak_data.uno.death_event)

		self._post_death_t = t + tweak_data.uno.post_death_timer
		self._death_t = nil
	end

	if self._post_death_t and t >= self._post_death_t then
		self:remove_controllers()

		managers.network:session():send_to_peers("mission_ended", false, 0)
		game_state_machine:change_state_by_name("gameoverscreen")

		self._force_feed_endscreen_statistics = true
		self._post_death_t = nil
	end

	if self._force_feed_endscreen_statistics then
		local count = self:player_count()
		managers.menu_component:feed_endscreen_statistics({
			best_killer = managers.localization:text("victory_best_killer_name", {
				PLAYER_NAME = "Me",
				SCORE = count
			}),
			best_special = managers.localization:text("victory_best_special_name", {
				PLAYER_NAME = "Me",
				SCORE = count
			}),
			best_accuracy = managers.localization:text("victory_best_accuracy_name", {
				PLAYER_NAME = "Me",
				SCORE = "100"
			}),
			most_downs = managers.localization:text("victory_most_downs_name", {
				PLAYER_NAME = "Everyone",
				SCORE = count
			}),
			total_kills = count,
			total_specials_kills = count,
			total_head_shots = count,
			group_hit_accuracy = "100%",
			group_total_downed = count,
			stage_cash_summary = managers.localization:text("heist_just_uno_end")
		})
	end

	if Network:is_server() then
		if self._next_card_draw_t and self._next_card_draw_t <= t then
			local player = ((self._cards_to_distribute - 1) % self:player_count() + 1)
			self:_server_draw_card(player, false)

			self._cards_to_distribute = self._cards_to_distribute - 1

			if self._cards_to_distribute > 0 then
				self._next_card_draw_t = t + tweak_data.uno.intro_card_cooldown
			else
				self._next_card_draw_t = nil
				self._cards_to_distribute = nil

				self:set_turn(1)
				-- self:set_turn(math.random(self:player_count()))
				managers.uno_network:send_peers_request("SYNC_TURN", {self._turn})
			end
		end

		if self._turn > 0 then
			local peers_table = managers.network and managers.network:session() and managers.network:session():all_peers()
			for player_id, seat_id in pairs(self._player_to_seat) do
				if not peers_table[player_id] then
					self:_handle_ai_control(t, dt, player_id)
				end

				local hand = self._hands[seat_id]
				if hand:card_amount() == 0 then
					self:set_turn(0)
					managers.uno_network:send_peers_request("SYNC_TURN", {self._turn})

					managers.uno_network:send_peers_request("SYNC_WINNER", {player_id})
				end
			end
		end
	end
end

function UNOManager:_draw_blank_card(recipient_id)
	local card = self:spawn_card()
	local hand = self._hands[recipient_id]

	card:set_owner(hand)
	managers.menu_component:post_event("loot_flip_card")
end

function UNOManager:_draw_card_local(card_id)
	local card = self:spawn_card()
	local hand = self:local_hand()

	card:set_card_id(card_id)

	card:set_owner(hand)
	managers.menu_component:post_event("loot_flip_card")
end

function UNOManager:play_card_local(card, card_index)
	card_index = card_index or self:local_hand():card_amount()

	card:set_owner(self._play_pile)

	local default_behaviour_overriden = false
	local local_placement_clbk = card:local_placement_clbk()
	if local_placement_clbk then
		default_behaviour_overriden = local_placement_clbk(card, card_index)
	end

	if not default_behaviour_overriden then
		managers.uno_network:send_host_request("REQUEST_PLAY_CARD", {card:card_id(), card_index})
		managers.uno_network:send_host_request("END_TURN")
	end

	managers.menu_component:post_event("loot_flip_card")
end

function UNOManager:_play_card(card, card_id)
	card = card or self:spawn_card()
	if card_id then
		card:set_card_id(card_id)
	end

	card:set_owner(self._play_pile)
	managers.menu_component:post_event("loot_flip_card")
end

function UNOManager:_register_network_messages()
	local function sync_intro_start(sender_id, response_identifier)
		game_state_machine:change_state_by_name("ingame_uno")

		self._intro_started = true
	end
	managers.uno_network:register_network_message("SYNC_INTRO_START", sync_intro_start)

	local function sync_draw_card(sender_id, response_identifier, recipient_id)
		self:_draw_blank_card(recipient_id)
	end
	managers.uno_network:register_network_message("SYNC_DRAW_CARD", sync_draw_card, {
		UNONetworkManager.ARGUMENT_TYPES.NUMBER
	})

	local function client_give_card(sender_id, response_identifier, card_id)
		card_id = tweak_data.uno.from_obfuscated[card_id]
		self:_draw_card_local(card_id)
	end
	managers.uno_network:register_network_message("CLIENT_GIVE_CARD", client_give_card, {
		UNONetworkManager.ARGUMENT_TYPES.STRING
	})

	local function client_force_play(sender_id, response_identifier, card_id)
		local card = self:spawn_card()
		card:set_card_id(card_id)

		self:play_card_local(card)
	end
	managers.uno_network:register_network_message("CLIENT_FORCE_PLAY", client_force_play, {
		UNONetworkManager.ARGUMENT_TYPES.STRING
	})

	local function request_draw_card(sender_id, response_identifier)
		-- Do validation or something?

		if self:_server_draw_card(sender_id, true) then
			self:increment_turn()
			managers.uno_network:send_peers_request("SYNC_TURN", {self._turn})
		end
	end
	managers.uno_network:register_network_message("REQUEST_DRAW_CARD", request_draw_card)

	local function sync_play_card(sender_id, response_identifier, playing_id, card_id, index)
		local hand = self._hands[playing_id]
		local card = hand:get_card(index)

		self:_play_card(card, card_id)
	end
	managers.uno_network:register_network_message("SYNC_PLAY_CARD", sync_play_card, {
		UNONetworkManager.ARGUMENT_TYPES.NUMBER,
		UNONetworkManager.ARGUMENT_TYPES.STRING,
		UNONetworkManager.ARGUMENT_TYPES.NUMBER
	})

	local function request_play_card(sender_id, response_identifier, card_id, index)
		-- Do validation or something?

		local card_tweak_data = tweak_data.uno.cards[card_id]
		if card_tweak_data.server_placement_clbk then
			card_tweak_data.server_placement_clbk()
		end

		managers.uno_network:send_peers_except_request(sender_id, "SYNC_PLAY_CARD", {sender_id, card_id, index})
	end
	managers.uno_network:register_network_message("REQUEST_PLAY_CARD", request_play_card, {
		UNONetworkManager.ARGUMENT_TYPES.STRING,
		UNONetworkManager.ARGUMENT_TYPES.NUMBER
	})

	local function sync_turn(sender_id, response_identifier, turn)
		self:set_turn(turn)

		for _, hand in pairs(self._hands) do
			hand:update_cards()
		end
	end
	managers.uno_network:register_network_message("SYNC_TURN", sync_turn, {
		UNONetworkManager.ARGUMENT_TYPES.NUMBER
	})

	local function sync_winner(sender_id, response_identifier, winner_id)
		managers.overlay_effect:play_effect({
			blend_mode = "normal",
			sustain = tweak_data.uno.post_death_timer + tweak_data.uno.death_timer,
			fade_out = 1,
			fade_in = 1,
			play_paused = false,
			color = Color(1, 0, 0, 0),
			timer = TimerManager:main()
		})

		self._game_over = true
	end
	managers.uno_network:register_network_message("SYNC_WINNER", sync_winner)

	local function request_end_turn(sender_id, response_identifier)
		if self:is_turn(sender_id) then
			self:increment_turn()

			managers.uno_network:send_peers_request("SYNC_TURN", {self._turn})
		end
	end
	managers.uno_network:register_network_message("END_TURN", request_end_turn)

	local function sync_card_color(sender_id, response_identifier, color_id)
		if not self._play_pile:top_card() then return end

		self._play_pile:top_card():set_type(color_id)
	end
	managers.uno_network:register_network_message("SYNC_CARD_COLOR", sync_card_color, {
		UNONetworkManager.ARGUMENT_TYPES.STRING
	})

	local function request_card_color(sender_id, response_identifier, color_id)
		if self:is_turn(sender_id) then
			managers.uno_network:send_peers_request("SYNC_CARD_COLOR", {color_id})
		end
	end
	managers.uno_network:register_network_message("REQUEST_CARD_COLOR", request_card_color, {
		UNONetworkManager.ARGUMENT_TYPES.STRING
	})

	local function sync_ai_character(sender_id, response_identifier, player_id, character_name, henchman_string)
		local character_location = self._arena_info.characters[player_id]
		local henchman_loadout = managers.blackmarket:unpack_henchman_loadout_string(henchman_string)
		local character_id = managers.blackmarket:get_character_id_by_character_name(character_name)

		local complete_outfit = {
			character = character_name,
			mask = {
				mask_id = henchman_loadout.mask,
				blueprint = henchman_loadout.mask_blueprint,
			},
			armor = "level_1",
			armor_current = "level_1",
			armor_skin = "none",
			player_style = henchman_loadout.player_style,
			suit_variation = henchman_loadout.suit_variation,
			glove_id = henchman_loadout.glove_id
		}

		self._characters[player_id] = UNOCharacter:new(character_location[1], character_location[2], character_id, complete_outfit)
	end
	managers.uno_network:register_network_message("SYNC_AI_CHARACTER", sync_ai_character, {
		UNONetworkManager.ARGUMENT_TYPES.NUMBER,
		UNONetworkManager.ARGUMENT_TYPES.STRING,
		UNONetworkManager.ARGUMENT_TYPES.STRING
	})
end

Hooks:PostHook(GameSetup, "init_managers", "InitUNOManager", function(self)
	managers.uno = UNOManager:new()
end)

Hooks:PostHook(GameSetup, "update", "UpdateUNOManager", function(self, t, dt)
	managers.uno:update(t, dt)
end)