UNONetworkManager = UNONetworkManager or class()
UNONetworkManager.ARGUMENT_TYPES = {
	STRING = 0,
	NUMBER = 1,
	BOOLEAN = 2
}

function UNONetworkManager:init()
	self._network_message_arguments = {}
	self._network_message_receive_functions = {}

	self._network_queue = {}

	self._responses = {}
	self._awaiting_response = {}
end

function UNONetworkManager:register_network_message(message_id, receive_function, arguments)
	self._network_message_arguments[message_id] = arguments and clone(arguments) or {}
	self._network_message_receive_functions[message_id] = receive_function

	self._awaiting_response[message_id] = {}
end

function UNONetworkManager:update(t, dt)
	function timestamp_sort(a, b)
		return a[1] < b[1]
	end
	table.sort(self._responses, timestamp_sort)
	table.sort(self._network_queue, timestamp_sort)

	for _, response_data in pairs(self._responses) do
		local sender_id = response_data[2]
		local response_identifier = response_data[3]
		local message_id = response_data[4]
		local response_argument = response_data[5]

		local response_clbk = self._awaiting_response[message_id][response_identifier]
		if response_clbk then
			response_clbk(response_argument)
		end

		self._awaiting_response[message_id][response_identifier] = nil
	end
	self._responses = {}

	for _, message_data in pairs(self._network_queue) do
		local sender_id = message_data[2]
		local response_identifier = message_data[3]
		local message_id = message_data[4]
		local argument_string = message_data[5]

		local argument_data = self._network_message_arguments[message_id]
		local receive_function = self._network_message_receive_functions[message_id]

		if argument_data and receive_function then
			local arguments = self:_decode_arguments(argument_string or "", argument_data)
			local response = receive_function(sender_id, response_identifier, unpack(arguments))

			if response then
				managers.uno_network:send_response(sender_id, message_id, response_identifier, response)
			end
		end
	end
	self._network_queue = {}
end

function UNONetworkManager:_encode_arguments(arguments)
	local argument_string = ""

	for index, argument in pairs(arguments) do
		argument_string = argument_string .. tostring(argument)

		if index ~= #arguments then
			argument_string = argument_string .. "|"
		end
	end

	return argument_string
end

function UNONetworkManager:_decode_arguments(argument_string, argument_data)
	local arguments = string.split(argument_string, "|")

	for index, argument in pairs(arguments) do
		local argument_type = argument_data[index]

		if argument_type then
			if argument_type == UNONetworkManager.ARGUMENT_TYPES.NUMBER then
				arguments[index] = tonumber(argument)
			elseif argument_type == UNONetworkManager.ARGUMENT_TYPES.BOOLEAN then
				arguments[index] = argument == "true"
			end
		end
	end

	return arguments
end

function UNONetworkManager:_receive_message(sender_id, message_id, data)
	local message_id_data = string.split(message_id, "|")

	local is_uno = message_id_data[1] == "UNO"
	if not is_uno then return end

	local is_response = message_id_data[2] == "RESPONSE"
	local timestamp = tonumber(message_id_data[3])
	local identifier = message_id_data[4]
	local message_id = message_id_data[5]

	if is_response then
		table.insert(self._responses, {timestamp, sender_id, identifier, message_id, data or ""})
	else
		table.insert(self._network_queue, {timestamp, sender_id, identifier, message_id, data or ""})
	end
end

function UNONetworkManager:_append_response_header(is_response, message_id, identifier)
	local timestamp = tostring(os.clock()) -- Technically relative to the Lua states start, but this is mostly here to keep short bursts of single user messages in order.
	if is_response then
		return "UNO|RESPONSE|" .. timestamp .. "|" .. identifier .. "|" .. message_id
	else
		return "UNO|REQUEST|" .. timestamp .. "|" .. identifier .. "|" .. message_id
	end
end

function UNONetworkManager:send_response(sender_id, response_id, identifier, response)
	if not sender_id then return end
	if not response_id then return end
	if not identifier then return end
	if not response then return end

	response_id = self:_append_response_header(true, response_id, identifier)
	LuaNetworking:SendToPeer(sender_id, response_id, tostring(response))
end

function UNONetworkManager:_contains_peer_id(id, peer_id)
	local id_list = type(id) == "list" and id or {id}
	if table.contains(id_list, peer_id) then
		return true
	end

	return false
end

local function generate_identifier()
	local template = "xxxxxx"
	return string.gsub(template, '[xy]', function (c)
		local v = (c == 'x') and math.random(0, 0xf) or math.random(8, 0xb)
		return string.format('%x', v)
	end)
end

function UNONetworkManager:_setup_response(message_id, response_clbk)
	local identifier = generate_identifier()
	if response_clbk and self._awaiting_response[message_id] then
		self._awaiting_response[message_id][identifier] = response_clbk
	end

	return identifier
end

function UNONetworkManager:send_peers_request(message_id, arguments, response_clbk)
	local identifier = self:_setup_response(message_id, response_clbk)
	message_id = self:_append_response_header(false, message_id, identifier)
	local encoded_arguments = self:_encode_arguments(arguments or {})

	LuaNetworking:SendToPeers(message_id, encoded_arguments)

	-- Receive our own message.
	local local_peer_id = managers.network:session():local_peer():id()
	self:_receive_message(local_peer_id, message_id, encoded_arguments)
end

function UNONetworkManager:send_peers_except_request(exception, message_id, arguments, response_clbk)
	local identifier = self:_setup_response(message_id, response_clbk)
	message_id = self:_append_response_header(false, message_id, identifier)
	local encoded_arguments = self:_encode_arguments(arguments or {})

	LuaNetworking:SendToPeersExcept(exception, message_id, encoded_arguments)

	-- Receive our own message.
	local local_peer_id = managers.network:session():local_peer():id()
	if not self:_contains_peer_id(exception, local_peer_id) then
		self:_receive_message(local_peer_id, message_id, encoded_arguments)
	end
end

function UNONetworkManager:send_peer_request(recipient_id, message_id, arguments, response_clbk)
	local identifier = self:_setup_response(message_id, response_clbk)
	message_id = self:_append_response_header(false, message_id, identifier)
	local encoded_arguments = self:_encode_arguments(arguments or {})

	LuaNetworking:SendToPeer(recipient_id, message_id, encoded_arguments)

	local local_peer_id = managers.network:session():local_peer():id()
	if local_peer_id == recipient_id then -- If we are the recipient then receive our own message.
		self:_receive_message(recipient_id, message_id, encoded_arguments)
	end
end

function UNONetworkManager:send_host_request(message_id, arguments, response_clbk)
	local identifier = self:_setup_response(message_id, response_clbk)
	message_id = self:_append_response_header(false, message_id, identifier)
	local encoded_arguments = self:_encode_arguments(arguments or {})

	LuaNetworking:SendToPeer(1, message_id, encoded_arguments)

	local local_peer_id = managers.network:session():local_peer():id()
	if local_peer_id == 1 then -- If we are the host then receive our own message.
		self:_receive_message(1, message_id, encoded_arguments)
	end
end

Hooks:Add("NetworkReceivedData", "NetworkReceivedData_UNONetworkManager", function(sender_id, message_id, data)
	if not managers.uno_network then return end

	managers.uno_network:_receive_message(sender_id, message_id, data)
end)

Hooks:PostHook(GameSetup, "init_managers", "InitUNONetworkManager", function(self)
	managers.uno_network = UNONetworkManager:new()
end)

Hooks:PostHook(GameSetup, "update", "UpdateUNONetworkManager", function(self, t, dt)
	managers.uno_network:update(t, dt)
end)